package ca.recrown.islandsurvivalcraft.world.generation;


import org.bukkit.Chunk;
import org.bukkit.World;
import org.bukkit.block.Biome;

import ca.recrown.islandsurvivalcraft.caching.Cache;
import ca.recrown.islandsurvivalcraft.caching.CacheValue;
import ca.recrown.islandsurvivalcraft.caching.CoordinateIdentifier;
import ca.recrown.islandsurvivalcraft.pathfinding.CoordinateTargetValidatable;
import ca.recrown.islandsurvivalcraft.pathfinding.CoordinateValidatable;
import ca.recrown.islandsurvivalcraft.pathfinding.DepthFirstSearch;
import ca.recrown.islandsurvivalcraft.world.BiomeSelector;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMapper;

//Note: technically, the validators have to be run on land, and so, some condition checks may not be nessecary.
public class BiomePerIslandGenerator implements IslandBiomeGenerator {
    private boolean initialized;
    private final TemperatureMapGenerator temperatureMapGenerator;
    private final Cache<Biome[][]> chunkBiomesCache;
    private final Cache<Boolean> chunkGenStatusCache;
    private IslandWorldMapper worldIslandMap;
    private BiomeSelector biomeSelector;
    private World world;
    private final DepthFirstSearch freshCachePropagator;
    private final DepthFirstSearch existenceChecker;
    private FreshCachePropagationInfo freshCachePropInfo;
    private PreviousGenerationInfo existenceInfo;
    CoordinateIdentifier currChunkCoords;
    
    float temperature;

    public BiomePerIslandGenerator() {
        this.temperatureMapGenerator = new TemperatureMapGenerator();
        chunkBiomesCache = new Cache<>(1024);
        chunkGenStatusCache = new Cache<>(1024);
        freshCachePropInfo = new FreshCachePropagationInfo();
        freshCachePropagator = new DepthFirstSearch(freshCachePropInfo);
        existenceInfo = new PreviousGenerationInfo();
        existenceChecker = new DepthFirstSearch(existenceInfo);
    }
    
    @Override
    public void initialize(final World world, final IslandWorldMapper mapGenerator, final BiomeSelector biomeSelector) {
        if (initialized) throw new IllegalStateException("Biome generator already initialized.");
        initialized = true;
        this.world = world;
        this.worldIslandMap = mapGenerator;
        this.biomeSelector = biomeSelector;
        this.temperatureMapGenerator.setSeed(world.getSeed());
    }

    @Override
    public IslandBiomeGenerator getInstance() {
        return new BiomePerIslandGenerator();
    }

    @Override
    public Biome GenerateBiome(int chunkX, int chunkZ, int localX, int localZ) {
        if (currChunkCoords == null || chunkX != currChunkCoords.getX() || chunkZ != currChunkCoords.getY()) {
            currChunkCoords = new CoordinateIdentifier(chunkX, chunkZ);
        }
        int worldX = localX + 16 * chunkX;
        int worldZ = localZ + 16 * chunkZ;

        Biome cachedBiome = getBiome(worldX, worldZ);
        if (cachedBiome != null) return cachedBiome;
        temperature = temperatureMapGenerator.getTemperature(worldX, worldZ);

        if (!worldIslandMap.isIsland(worldX, worldZ)) {
            return biomeSelector.getOceanBiome(temperature);
        }
        
        freshCachePropInfo.clear();
        existenceChecker.setStartPosition(worldX, worldZ);
        if (!existenceChecker.findTarget(existenceInfo)) {
            if (freshCachePropInfo.main == null) freshCachePropInfo.main = biomeSelector.getLandBiome(temperature);
            if (freshCachePropInfo.shore == null) freshCachePropInfo.shore = biomeSelector.getShoreBiome(freshCachePropInfo.main, temperature);
            if (freshCachePropInfo.shallow == null) freshCachePropInfo.shallow = biomeSelector.getOceanBiome(temperature);            
        }
        freshCachePropagator.setStartPosition(worldX, worldZ);
        freshCachePropagator.findTarget(freshCachePropInfo);
        return getBiome(worldX, worldZ);
    }

    private Biome getBiome(int worldX, int worldZ) {
        int localX = Math.abs(worldX % 16);
        int localZ = Math.abs(worldZ % 16);
        CoordinateIdentifier chunkCoords = new CoordinateIdentifier(worldX / 16, worldZ / 16);

        CacheValue<Biome[][]> chunkBiomes = chunkBiomesCache.retrieveCache(chunkCoords);
        if (!chunkBiomes.isEmpty()) {
            Biome biome = chunkBiomes.getValue()[localX][localZ];
            if (biome != null) return biome;
        }

        CacheValue<Boolean> chunkGenStatus = chunkGenStatusCache.retrieveCache(chunkCoords);
        if (chunkGenStatus.isEmpty())
            chunkGenStatus.setValue(world.isChunkGenerated(chunkCoords.getX(), chunkCoords.getY()));
        
        if (chunkGenStatus.getValue()) {
            if (chunkBiomes.isEmpty()) chunkBiomes.setValue(new Biome[16][16]);
            chunkBiomes.getValue()[localX][localZ] = world.getBiome(worldX, 0, worldZ);
            return chunkBiomes.getValue()[localX][localZ];
        }
        return null;
    }

    private void setCacheBiome(int worldX, int worldZ, Biome biome) {
        int localX = Math.abs(worldX % 16);
        int localZ = Math.abs(worldZ % 16);
        CoordinateIdentifier chunkCoords = new CoordinateIdentifier(worldX / 16, worldZ / 16);

        CacheValue<Biome[][]> chunkBiomes = chunkBiomesCache.retrieveCache(chunkCoords);
        if (chunkBiomes.isEmpty()) chunkBiomes.setValue(new Biome[16][16]);
        chunkBiomes.getValue()[localX][localZ] = biome;
    }

    private class FreshCachePropagationInfo implements CoordinateTargetValidatable, CoordinateValidatable {
        public Biome shore;
        public Biome main;
        public Biome shallow;

        @Override
        public boolean isCoordinateTarget(int x, int y) {
            if (worldIslandMap.isLand(x, y)) {
                if (worldIslandMap.isShore(x, y)) {
                    setCacheBiome(x, y, shore);
                } else {
                    setCacheBiome(x, y, main);
                }
            } else {
                setCacheBiome(x, y, shallow);
            }
            return false;
        }
        
        @Override
        public boolean validate(int x, int y) {
            return x / 16 == currChunkCoords.getX() && y / 16 == currChunkCoords.getY() && worldIslandMap.isIsland(x, y);
        }

        public boolean allBiomesAcquired() {
            return shore != null && main != null && shallow != null;
        }

        public void clear() {
            main = null;
            shore = null;
            shallow = null;
        }
    }

    private class PreviousGenerationInfo implements CoordinateTargetValidatable, CoordinateValidatable {

        @Override
        public boolean validate(int x, int y) {
            return worldIslandMap.isIsland(x, y);
        }

        @Override
        public boolean isCoordinateTarget(int x, int y) {
            if (freshCachePropInfo.main == null && worldIslandMap.isLand(x, y) && !worldIslandMap.isShore(x, y)) {
                freshCachePropInfo.main = getBiome(x, y);
            }
            if (freshCachePropInfo.shore == null && worldIslandMap.isShore(x, y)) {
                freshCachePropInfo.shore = getBiome(x, y);
            }
            if (freshCachePropInfo.shallow == null && worldIslandMap.isShallowPortion(x, y)) {
                freshCachePropInfo.shallow = getBiome(x, y);
            }
            return freshCachePropInfo.allBiomesAcquired();
        }

    }
}